@@ -1,5 +1,5 @@
 /*
- * $RCSId: xc/lib/fontconfig/src/fcstr.c,v 1.10 2002/08/31 22:17:32 keithp Exp $
+ * fontconfig/src/fcstr.c
  *
  * Copyright © 2000 Keith Packard
  *
@@ -7,15 +7,15 @@
  * documentation for any purpose is hereby granted without fee, provided that
  * the above copyright notice appear in all copies and that both that
  * copyright notice and this permission notice appear in supporting
- * documentation, and that the name of Keith Packard not be used in
+ * documentation, and that the name of the author(s) not be used in
  * advertising or publicity pertaining to distribution of the software without
- * specific, written prior permission.  Keith Packard makes no
+ * specific, written prior permission.  The authors make no
  * representations about the suitability of this software for any purpose.  It
  * is provided "as is" without express or implied warranty.
  *
- * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
+ * THE AUTHOR(S) DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
  * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
- * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
+ * EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY SPECIAL, INDIRECT OR
  * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
  * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
  * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
@@ -26,6 +26,9 @@
 #include <stdlib.h>
 #include <ctype.h>
 #include <string.h>
+#ifdef HAVE_REGEX_H
+#include <regex.h>
+#endif
 #ifdef _WIN32
 #include <windows.h>
 #endif
@@ -50,14 +53,16 @@ FcStrCopy (const FcChar8 *s)
 FcChar8 *
 FcStrPlus (const FcChar8 *s1, const FcChar8 *s2)
 {
-    int	    l = strlen ((char *)s1) + strlen ((char *) s2) + 1;
+    int	    s1l = strlen ((char *) s1);
+    int	    s2l = strlen ((char *) s2);
+    int	    l = s1l + s2l + 1;
     FcChar8 *s = malloc (l);
 
     if (!s)
 	return 0;
     FcMemAlloc (FC_MEM_STRING, l);
-    strcpy ((char *) s, (char *) s1);
-    strcat ((char *) s, (char *) s2);
+    memcpy (s, s1, s1l);
+    memcpy (s + s1l, s2, s2l + 1);
     return s;
 }
 
@@ -109,7 +114,7 @@ FcStrCaseWalkerLong (FcCaseWalker *w, FcChar8 r)
 	    int		mid = (min + max) >> 1;
 	    FcChar32    low = fcCaseFold[mid].upper;
 	    FcChar32    high = low + FcCaseFoldUpperCount (&fcCaseFold[mid]);
-	    
+	
 	    if (high <= ucs4)
 		min = mid + 1;
 	    else if (ucs4 < low)
@@ -158,7 +163,7 @@ FcStrCaseWalkerNext (FcCaseWalker *w)
 	w->read = 0;
     }
     r = *w->src++;
-    
+
     if ((r & 0xc0) == 0xc0)
 	return FcStrCaseWalkerLong (w, r);
     if ('A' <= r && r <= 'Z')
@@ -181,7 +186,7 @@ FcStrCaseWalkerNextIgnoreBlanks (FcCaseWalker *w)
     {
 	r = *w->src++;
     } while (r == ' ');
-    
+
     if ((r & 0xc0) == 0xc0)
 	return FcStrCaseWalkerLong (w, r);
     if ('A' <= r && r <= 'Z')
@@ -215,11 +220,11 @@ FcStrCmpIgnoreCase (const FcChar8 *s1, const FcChar8 *s2)
     FcChar8	    c1, c2;
 
     if (s1 == s2) return 0;
-    
+
     FcStrCaseWalkerInit (s1, &w1);
     FcStrCaseWalkerInit (s2, &w2);
-    
-    for (;;) 
+
+    for (;;)
     {
 	c1 = FcStrCaseWalkerNext (&w1);
 	c2 = FcStrCaseWalkerNext (&w2);
@@ -236,11 +241,11 @@ FcStrCmpIgnoreBlanksAndCase (const FcChar8 *s1, const FcChar8 *s2)
     FcChar8	    c1, c2;
 
     if (s1 == s2) return 0;
-    
+
     FcStrCaseWalkerInit (s1, &w1);
     FcStrCaseWalkerInit (s2, &w2);
-    
-    for (;;) 
+
+    for (;;)
     {
 	c1 = FcStrCaseWalkerNextIgnoreBlanks (&w1);
 	c2 = FcStrCaseWalkerNextIgnoreBlanks (&w2);
@@ -254,10 +259,10 @@ int
 FcStrCmp (const FcChar8 *s1, const FcChar8 *s2)
 {
     FcChar8 c1, c2;
-    
+
     if (s1 == s2)
 	return 0;
-    for (;;) 
+    for (;;)
     {
 	c1 = *s1++;
 	c2 = *s2++;
@@ -267,6 +272,55 @@ FcStrCmp (const FcChar8 *s1, const FcChar8 *s2)
     return (int) c1 - (int) c2;
 }
 
+#ifdef USE_REGEX
+static FcBool
+_FcStrRegexCmp (const FcChar8 *s, const FcChar8 *regex, int cflags, int eflags)
+{
+    int ret = -1;
+    regex_t reg;
+
+    if ((ret = regcomp (&reg, (const char *)regex, cflags)) != 0)
+    {
+	if (FcDebug () & FC_DBG_MATCHV)
+	{
+	    char buf[512];
+
+	    regerror (ret, &reg, buf, 512);
+	    printf("Regexp compile error: %s\n", buf);
+	}
+	return FcFalse;
+    }
+    ret = regexec (&reg, (const char *)s, 0, NULL, eflags);
+    if (ret != 0)
+    {
+	if (FcDebug () & FC_DBG_MATCHV)
+	{
+	    char buf[512];
+
+	    regerror (ret, &reg, buf, 512);
+	    printf("Regexp exec error: %s\n", buf);
+	}
+    }
+    regfree (&reg);
+
+    return ret == 0 ? FcTrue : FcFalse;
+}
+#else
+#  define _FcStrRegexCmp(_s_, _regex_, _cflags_, _eflags_)	(FcFalse)
+#endif
+
+FcBool
+FcStrRegexCmp (const FcChar8 *s, const FcChar8 *regex)
+{
+	return _FcStrRegexCmp (s, regex, REG_EXTENDED | REG_NOSUB, 0);
+}
+
+FcBool
+FcStrRegexCmpIgnoreCase (const FcChar8 *s, const FcChar8 *regex)
+{
+	return _FcStrRegexCmp (s, regex, REG_EXTENDED | REG_NOSUB | REG_ICASE, 0);
+}
+
 /*
  * Return a hash value for a string
  */
@@ -296,8 +350,8 @@ FcStrIsAtIgnoreBlanksAndCase (const FcChar8 *s1, const FcChar8 *s2)
 
     FcStrCaseWalkerInit (s1, &w1);
     FcStrCaseWalkerInit (s2, &w2);
-    
-    for (;;) 
+
+    for (;;)
     {
 	c1 = FcStrCaseWalkerNextIgnoreBlanks (&w1);
 	c2 = FcStrCaseWalkerNextIgnoreBlanks (&w2);
@@ -355,8 +409,8 @@ FcStrIsAtIgnoreCase (const FcChar8 *s1, const FcChar8 *s2)
 
     FcStrCaseWalkerInit (s1, &w1);
     FcStrCaseWalkerInit (s2, &w2);
-    
-    for (;;) 
+
+    for (;;)
     {
 	c1 = FcStrCaseWalkerNext (&w1);
 	c2 = FcStrCaseWalkerNext (&w2);
@@ -395,7 +449,7 @@ FcStrContainsWord (const FcChar8 *s1, const FcChar8 *s2)
 	
     while (s1len >= s2len)
     {
-	if (wordStart && 
+	if (wordStart &&
 	    FcStrIsAtIgnoreCase (s1, s2) &&
 	    (s1len == s2len || FcCharIsPunct (s1[s2len])))
 	{
@@ -422,12 +476,12 @@ FcStrStrIgnoreCase (const FcChar8 *s1, const FcChar8 *s2)
 
     if (s1 == s2)
 	return s1;
-    
+
     FcStrCaseWalkerInit (s1, &w1);
     FcStrCaseWalkerInit (s2, &w2);
-    
+
     c2 = FcStrCaseWalkerNext (&w2);
-    
+
     for (;;)
     {
 	cur = w1.src;
@@ -474,7 +528,7 @@ again:
     if (!c2)
 	return 0;
 
-    for (;;) 
+    for (;;)
     {
 	p = s1;
 	c1 = *s1++;
@@ -517,15 +571,15 @@ FcUtf8ToUcs4 (const FcChar8 *src_orig,
 
     if (len == 0)
 	return 0;
-    
+
     s = *src++;
     len--;
-    
+
     if (!(s & 0x80))
     {
 	result = s;
 	extra = 0;
-    } 
+    }
     else if (!(s & 0x40))
     {
 	return -1;
@@ -561,7 +615,7 @@ FcUtf8ToUcs4 (const FcChar8 *src_orig,
     }
     if (extra > len)
 	return -1;
-    
+
     while (extra--)
     {
 	result <<= 6;
@@ -586,7 +640,7 @@ FcUtf8Len (const FcChar8    *string,
     int		clen;
     FcChar32	c;
     FcChar32	max;
-    
+
     n = 0;
     max = 0;
     while (len)
@@ -616,7 +670,7 @@ FcUcs4ToUtf8 (FcChar32	ucs4,
 {
     int	bits;
     FcChar8 *d = dest;
-    
+
     if      (ucs4 <       0x80) {  *d++=  ucs4;                         bits= -6; }
     else if (ucs4 <      0x800) {  *d++= ((ucs4 >>  6) & 0x1F) | 0xC0;  bits=  0; }
     else if (ucs4 <    0x10000) {  *d++= ((ucs4 >> 12) & 0x0F) | 0xE0;  bits=  6; }
@@ -647,11 +701,11 @@ FcUtf16ToUcs4 (const FcChar8	*src_orig,
 
     if (len < 2)
 	return 0;
-    
+
     a = GetUtf16 (src, endian); src += 2; len -= 2;
-    
-    /* 
-     * Check for surrogate 
+
+    /*
+     * Check for surrogate
      */
     if ((a & 0xfc00) == 0xd800)
     {
@@ -683,7 +737,7 @@ FcUtf16Len (const FcChar8   *string,
     int		clen;
     FcChar32	c;
     FcChar32	max;
-    
+
     n = 0;
     max = 0;
     while (len)
@@ -710,11 +764,18 @@ FcUtf16Len (const FcChar8   *string,
 void
 FcStrBufInit (FcStrBuf *buf, FcChar8 *init, int size)
 {
-    buf->buf = init;
+    if (init)
+    {
+	buf->buf = init;
+	buf->size = size;
+    } else
+    {
+	buf->buf = buf->buf_static;
+	buf->size = sizeof (buf->buf_static);
+    }
     buf->allocated = FcFalse;
     buf->failed = FcFalse;
     buf->len = 0;
-    buf->size = size;
 }
 
 void
@@ -733,7 +794,10 @@ FcStrBufDone (FcStrBuf *buf)
 {
     FcChar8 *ret;
 
-    ret = malloc (buf->len + 1);
+    if (buf->failed)
+	ret = NULL;
+    else
+	ret = malloc (buf->len + 1);
     if (ret)
     {
 	FcMemAlloc (FC_MEM_STRING, buf->len + 1);
@@ -744,6 +808,17 @@ FcStrBufDone (FcStrBuf *buf)
     return ret;
 }
 
+FcChar8 *
+FcStrBufDoneStatic (FcStrBuf *buf)
+{
+    FcStrBufChar (buf, '\0');
+
+    if (buf->failed)
+	return NULL;
+
+    return buf->buf;
+}
+
 FcBool
 FcStrBufChar (FcStrBuf *buf, FcChar8 c)
 {
@@ -752,9 +827,13 @@ FcStrBufChar (FcStrBuf *buf, FcChar8 c)
 	FcChar8	    *new;
 	int	    size;
 
+	if (buf->failed)
+	    return FcFalse;
+
 	if (buf->allocated)
 	{
 	    size = buf->size * 2;
+	    FcMemFree (FC_MEM_STRBUF, buf->size);
 	    new = realloc (buf->buf, size);
 	}
 	else
@@ -772,8 +851,6 @@ FcStrBufChar (FcStrBuf *buf, FcChar8 c)
 	    buf->failed = FcTrue;
 	    return FcFalse;
 	}
-	if (buf->size)
-	    FcMemFree (FC_MEM_STRBUF, buf->size);
 	FcMemAlloc (FC_MEM_STRBUF, size);
 	buf->size = size;
 	buf->buf = new;
@@ -811,18 +888,18 @@ FcChar8 *
 FcStrCopyFilename (const FcChar8 *s)
 {
     FcChar8 *new;
-    
+
     if (*s == '~')
     {
 	FcChar8	*home = FcConfigHome ();
 	FcChar8	*full;
 	int	size;
 	if (!home)
-	    return 0;
+	    return NULL;
 	size = strlen ((char *) home) + strlen ((char *) s);
 	full = (FcChar8 *) malloc (size);
 	if (!full)
-	    return 0;
+	    return NULL;
 	strcpy ((char *) full, (char *) home);
 	strcat ((char *) full, (char *) s + 1);
 	new = FcStrCanonFilename (full);
@@ -830,6 +907,7 @@ FcStrCopyFilename (const FcChar8 *s)
     }
     else
 	new = FcStrCanonFilename (s);
+
     return new;
 }
 
@@ -851,7 +929,7 @@ FcStrLastSlash (const FcChar8  *path)
 
     return slash;
 }
-  
+
 FcChar8 *
 FcStrDirname (const FcChar8 *file)
 {
@@ -896,12 +974,19 @@ FcStrCanonAbsoluteFilename (const FcChar8 *s)
     FcMemAlloc (FC_MEM_STRING, size);
     slash = NULL;
     f = file;
+#ifdef _WIN32
+    if (*s == '/' && *(s+1) == '/') /* Network path, do not squash // */
+	*f++ = *s++;
+#endif
     for (;;) {
 	if (*s == '/' || *s == '\0')
 	{
 	    if (slash)
 	    {
 		switch (s - slash) {
+		case 1:
+		    f -= 1;	/* squash // and trim final / from file */
+		    break;
 		case 2:
 		    if (!strncmp ((char *) slash, "/.", 2))
 		    {
@@ -927,10 +1012,10 @@ FcStrCanonAbsoluteFilename (const FcChar8 *s)
     }
     return file;
 }
- 
+
 #ifdef _WIN32
 /*
- * Convert '\\' to '/' , remove double '/' 
+ * Convert '\\' to '/' , remove double '/'
  */
 static void
 FcConvertDosPath (char *str)
@@ -940,7 +1025,13 @@ FcConvertDosPath (char *str)
   char *dest = str;
   char *end = str + len;
   char last = 0;
-  
+
+  if (*p == '\\')
+    {
+      *p = '/';
+      p++;
+      dest++;
+    }
   while (p < end)
     {
       if (*p == '\\')
@@ -965,15 +1056,13 @@ FcStrCanonFilename (const FcChar8 *s)
 {
 #ifdef _WIN32
     FcChar8 full[FC_MAX_FILE_LEN + 2];
-    FcChar8 basename[FC_MAX_FILE_LEN + 2];
-    int size = GetFullPathName (s, sizeof (full) -1,
-				full,
-				basename);
+    int size = GetFullPathName ((LPCSTR) s, sizeof (full) -1,
+				(LPSTR) full, NULL);
 
     if (size == 0)
 	perror ("GetFullPathName");
 
-    FcConvertDosPath (full);
+    FcConvertDosPath ((char *) full);
     return FcStrCanonAbsoluteFilename (full);
 #else
     if (s[0] == '/')
@@ -1091,6 +1180,50 @@ FcStrSetAddFilename (FcStrSet *set, const FcChar8 *s)
     return FcTrue;
 }
 
+FcBool
+FcStrSetAddLangs (FcStrSet *strs, const char *languages)
+{
+    const char *p = languages, *next;
+    FcChar8 lang[128] = {0}, *normalized_lang;
+    size_t len;
+    FcBool ret = FcFalse;
+
+    if (!languages)
+	return FcFalse;
+
+    while ((next = strchr (p, ':')))
+    {
+	len = next - p;
+	len = FC_MIN (len, 128);
+	strncpy ((char *) lang, p, len);
+	lang[len] = 0;
+	/* ignore an empty item */
+	if (*lang)
+	{
+	    normalized_lang = FcLangNormalize ((const FcChar8 *) lang);
+	    if (normalized_lang)
+	    {
+		FcStrSetAdd (strs, normalized_lang);
+		free (normalized_lang);
+		ret = FcTrue;
+	    }
+	}
+	p = next + 1;
+    }
+    if (*p)
+    {
+	normalized_lang = FcLangNormalize ((const FcChar8 *) p);
+	if (normalized_lang)
+	{
+	    FcStrSetAdd (strs, normalized_lang);
+	    free (normalized_lang);
+	    ret = FcTrue;
+	}
+    }
+
+    return ret;
+}
+
 FcBool
 FcStrSetDel (FcStrSet *set, const FcChar8 *s)
 {
@@ -1104,7 +1237,7 @@ FcStrSetDel (FcStrSet *set, const FcChar8 *s)
 	     * copy remaining string pointers and trailing
 	     * NULL
 	     */
-	    memmove (&set->strs[i], &set->strs[i+1], 
+	    memmove (&set->strs[i], &set->strs[i+1],
 		     (set->num - i) * sizeof (FcChar8 *));
 	    set->num--;
 	    return FcTrue;
@@ -1118,7 +1251,7 @@ FcStrSetDestroy (FcStrSet *set)
     if (--set->ref == 0)
     {
 	int	i;
-    
+
 	for (i = 0; i < set->num; i++)
 	    FcStrFree (set->strs[i]);
 	if (set->strs)
